# pylint: disable=redefined-outer-name
"""
Test the vector data interpolators
"""
import pytest
import numpy as np
import numpy.testing as npt

from ..datasets.synthetic import CheckerBoard
from ..coordinates import grid_coordinates
from ..vector import VectorSpline2D
from ..utils import n_1d_arrays
from .utils import requires_numba


@pytest.fixture
def data2d():
    "Make 2 component vector data"
    synth = CheckerBoard(region=(2, 8, -3, 3), w_east=20, w_north=20)
    coordinates = grid_coordinates(synth.region, spacing=0.5)
    data = tuple(synth.predict(coordinates) for i in range(2))
    return coordinates, data


def test_vector2d(data2d):
    "See if the exact solution works"
    coords, data = data2d
    spline = VectorSpline2D().fit(coords, data)
    npt.assert_allclose(spline.score(coords, data), 1)
    npt.assert_allclose(spline.predict(coords), data, rtol=1e-3)
    # There should be 1 force per data point
    assert data[0].size == spline.force_coords_[0].size
    assert data[0].size * 2 == spline.force_.size
    npt.assert_allclose(spline.force_coords_, n_1d_arrays(coords, n=2))


def test_vector2d_twice(data2d):
    "Check that the force coordinates are updated if fitting a second time"
    coords, data = data2d
    spline = VectorSpline2D().fit(coords, data)
    npt.assert_allclose(spline.force_coords_, n_1d_arrays(coords, n=2))
    coords2 = tuple(i * 1000 for i in coords)
    spline.fit(coords2, data)
    npt.assert_allclose(spline.force_coords_, n_1d_arrays(coords2, n=2))
    assert not np.allclose(spline.force_coords_, n_1d_arrays(coords, n=2))


def test_vector2d_weights(data2d):
    "Use unit weights and a regular grid solution"
    coords, data = data2d
    weights = tuple(np.ones_like(data[0]) for i in range(2))
    spline = VectorSpline2D(shape=(11, 11)).fit(coords, data, weights)
    npt.assert_allclose(spline.score(coords, data), 1, rtol=1e-3)
    npt.assert_allclose(spline.predict(coords), data, rtol=5e-2)
    assert spline.force_coords_[0].size == 11 * 11
    assert spline.force_.size == 2 * 11 * 11


def test_vector2d_fails(data2d):
    "Should fail if not given 2 data components"
    coords, data = data2d
    spline = VectorSpline2D()
    with pytest.raises(ValueError):
        spline.fit(coords, data[0])
    with pytest.raises(ValueError):
        spline.fit(coords, data + coords)


@requires_numba
def test_vector2d_jacobian_implementations(data2d):
    "Compare the numba and numpy implementations."
    coords = data2d[0]
    jac_numpy = VectorSpline2D(engine="numpy").jacobian(coords, coords)
    jac_numba = VectorSpline2D(engine="numba").jacobian(coords, coords)
    npt.assert_allclose(jac_numpy, jac_numba)
